"""Test helpers for cs.aws_account."""
import warnings

from zope.component.testlayer import ZCMLFileLayer

import cs.aws_account


# see https://github.com/boto/boto3/issues/454
warnings.filterwarnings("ignore", message="unclosed.*<ssl.SSLSocket.*>")


AWS_ACCOUNT_INTEGRATION_LAYER = ZCMLFileLayer(cs.aws_account,
                                              zcml_file='ftesting.zcml',
                                              name='AWSAccountComponents')



"""Retry module for AWS throttling error."""
import time
import logging

# pylint: disable=global-variable-not-assigned, invalid-name


logger = logging.getLogger(__name__)


def aws_throttling_retry(max_retries=5, base=0.5, growth_factor=0.5):
    """Retry AWS throttling errors with exponential backoff time.

    Calculate time to sleep based on below function.
    The format is::
        base + (retries * growth_factor)

    Kwargs:
    max_retries: Max number of retries when hitting AWS throttling error
    base: base time for exponential backoff calculation
    growth_factor: growth factor used to calculate sleeping time for every retry
    """
    def decorator_retry(func):
        def wrapper(*args, **kwargs):  # pylint: disable=inconsistent-return-statements
            if max_retries < 1:
                raise ValueError("max_retries can't be less than 1")

            retries = 0
            while retries < max_retries:
                try:
                    return func(*args, **kwargs)
                except Exception as error:  # pylint: disable=broad-exception-caught
                    err_str = str(error).lower()
                    if 'throttling' in err_str or 'rate exceeded' in err_str:
                        retries += 1
                        if retries >= max_retries:
                            logger.warning("Received AWS throttling error. Exhausted all attempts.")
                            raise error
                        backoff_seconds = base + (retries * growth_factor)
                        logger.warning("Received AWS throttling error. Retrying in %s seconds...",
                                       backoff_seconds)
                        time.sleep(backoff_seconds)
                    else:
                        raise error
        return wrapper

    return decorator_retry



import os
import unittest

from zope import component

from ..account import Account
from ..interfaces import IRegionalAccounts
from ..regional_account import RegionalAccount
from ..regional_accounts import RegionalAccounts
from ..session import Session
from ..testing import AWS_ACCOUNT_INTEGRATION_LAYER


class IntegrationTestAWSAccountRegionalAccounts(unittest.TestCase):

    level = 2

    def setUp(self):
        self.session_kwargs = {
            'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
            'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY'),
        }
        self.session = Session(**self.session_kwargs)
        self.account = Account(self.session)
        self.ratelimit_kwargs = {'max_count': 1, 'interval': 1, 'block': False}
        self.filter = {'Partitions': {'aws': {'Regions': {'include': ['us-east-1']}}}}

        self.filter1_raccts = RegionalAccounts(
            self.ratelimit_kwargs, {'SessionParameters': self.session_kwargs}, self.filter)

        self.all_raccts = RegionalAccounts(
            self.ratelimit_kwargs, {'SessionParameters': self.session_kwargs})

    def test_account(self):
        self.assertEqual(self.filter1_raccts.account().account_id(), self.account.account_id())

    def test_getitem(self):
        self.assertTrue(isinstance(self.filter1_raccts['us-east-1'], RegionalAccount))
        with self.assertRaises(KeyError):
            self.filter1_raccts['bad']

    def test_get(self):
        self.assertTrue(isinstance(self.filter1_raccts.get('us-east-1'), RegionalAccount))
        self.assertIsNone(self.filter1_raccts.get('bad', None))

    def test_contains(self):
        self.assertTrue('us-east-1' in self.filter1_raccts)
        self.assertFalse('bad' in self.filter1_raccts)

    def test_keys(self):
        self.assertTrue('us-east-1' in self.filter1_raccts.keys())
        self.assertFalse('bad' in self.filter1_raccts.keys())

    def test_iter(self):
        self.assertEqual(list(self.filter1_raccts), ['us-east-1'])

    def test_values(self):
        self.assertEqual(len(self.filter1_raccts.values()), 1)
        self.assertTrue(isinstance(self.filter1_raccts.values()[0], RegionalAccount))

    def test_items(self):
        self.assertEqual(len(self.filter1_raccts.items()), 1)
        self.assertEqual(self.filter1_raccts.items()[0][0], 'us-east-1')
        self.assertTrue(isinstance(self.filter1_raccts.items()[0][1], RegionalAccount))

    def test_len(self):
        self.assertEqual(len(self.filter1_raccts), 1)

    def test_filter_application(self):
        self.assertGreater(len(self.all_raccts), len(self.filter1_raccts))
        self.assertIn('us-east-2', self.all_raccts)
        self.assertNotIn('us-east-2', self.filter1_raccts)

    def test_filter_mutation(self):
        self.assertIn('us-east-1', self.all_raccts)
        self.assertIn('us-east-2', self.all_raccts)
        self.all_raccts.filter['Partitions'] = {'aws': {'Regions': {'exclude': ['us-east-2']}}}
        self.assertNotIn('us-east-2', self.all_raccts)
        self.assertIn('us-east-1', self.all_raccts)


class IntegrationTestAWSAccountRegionalAccountsZCA(unittest.TestCase):

    layer = AWS_ACCOUNT_INTEGRATION_LAYER

    def setUp(self):
        # RateLimit, Account, Filter=None, RateLimitRegionSpec=None, service='ec2'
        self.kwargs = {
            'RateLimit': {'max_count': 1, 'interval': 1, 'block': False},
            'Account':  {
                'SessionParameters': {
                    'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
                    'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY'),
                },
            },
            'Filter': {
                'Partitions': {'aws': {'Regions': {'include': ['us-east-1']}}},
            },
        }

    def test_factories(self):
        # non-cached
        ra = component.createObject(u"cs.aws_account.regional_accounts", **self.kwargs)
        self.assertTrue(IRegionalAccounts.providedBy(ra))

        # cached
        self.assertIs(
                component.createObject(u"cs.aws_account.cached_regional_accounts",
                                       **self.kwargs),
                component.createObject(u"cs.aws_account.cached_regional_accounts",
                                       **self.kwargs)
            )



"""Provides custom exceptions for cs.aws_account."""


class AWSClientException(Exception):
    """Represents a boto3 error raised from making a call to AWS API."""

    def __init__(self, error, *_, **kwargs):
        """Prepare the error message from the given error."""
        self._error = error
        self._account = kwargs['AccountAlias']
        self._account_id = kwargs['AccountId']
        self._region = kwargs['Region']
        self._warning = "None"
        error_string = str(self._error)
        if "NoSuch" in error_string:
            self._warning = "Queried item does not exist. Check in the AWS Web Console."
        elif "AccessDenied" in error_string:
            self._warning = "Improper AWS IAM permissions. Verify application role has permission to access resource."
        elif ("AuthFailure" in error_string or
              "InvalidClientTokenId" in error_string or
              "UnrecognizedClientException" in error_string):
            self._warning = ("Attempted to access a non-existent resource. "
                             "Ensure the action, account, and region are removed from the RAS filter")
        self._error_message = (f"Error message: {error_string}. Account Alias: {str(self._account)}. "
                               f"Account ID: {str(self._account_id)}. "
                               f"Region: {str(self._region)}. Warning: {str(self._warning)}")
        super().__init__(self._error_message)

    def __str__(self):
        """Emit the prepared error message."""
        return str(self._error_message)



"""Interfaces for wiring together components using Boto3 sessions."""
from boto3.session import Session
from zope import interface
# pylint: disable=inherit-non-class


class IBoto3Session(interface.Interface):
    """Marker for a Boto3 session object."""


interface.classImplements(Session, IBoto3Session)



"""Components for interacting with AWS accounts."""
# pylint: disable=invalid-name
from threading import RLock
import operator

from cachetools import cached, cachedmethod, Cache
from zope import interface
from zope.component.factory import Factory
from botocore.config import Config

from .caching_key import aggregated_string_hash
from .interfaces import IAccount
from .session import session_factory
from .retry import aws_throttling_retry


@interface.implementer(IAccount)
class Account:
    """AWS account information.

    Args:
        session: cs.aws_account.Session instance
    """

    aws_client_config = Config(retries={"max_attempts": 10})

    def __init__(self, session):
        """Initialize the Account with a Read Lock."""
        self._session = session
        self._cache_aliases = Cache(maxsize=1)
        self._rlock = RLock()

    def account_id(self):
        """Return account identifier string."""
        return self._session.account_id()

    def alias(self):
        """Return the first available alias or else the account id."""
        aliases = self.aliases()
        return aliases[0] if aliases else self.account_id()

    @cachedmethod(operator.attrgetter('_cache_aliases'), lock=operator.attrgetter('_rlock'))
    @aws_throttling_retry()
    def aliases(self):
        """Return list of all account aliases."""
        client_kwargs = self.session().client_kwargs(service='iam')
        client_kwargs['config'] = Account.aws_client_config
        return self.session().boto3().client('iam', **client_kwargs).list_account_aliases()['AccountAliases']

    def session(self):
        """Return referenced cs.aws_account.Session object."""
        return self._session


AccountFactory = Factory(Account)


@interface.implementer(IAccount)
@cached(cache={}, key=aggregated_string_hash, lock=RLock())
def account_factory(SessionParameters=None, AssumeRole=None, AssumeRoles=None):
    """Create and cache a cs.aws_account.account.Account.

    Common call signatures will return cached object.

    Create cs.aws_account.account.Account object. If AssumeRole parameter is
    available, then process the role assumption. If AssumeRoles parameter
    is available, then process the series of role assumptions.

    Kwargs:
        SessionParameters: [see cs.aws_account.session.Session]
        AssumeRole: [see cs.aws_account.session.Session.assume_role]
        AssumeRoles: iterable of assume_role mappings

    Returns:
        cs.aws_account.account.Account object
    """
    session = session_factory(SessionParameters=SessionParameters,
                              AssumeRole=AssumeRole,
                              AssumeRoles=AssumeRoles)
    return Account(session=session)


CachingAccountFactory = Factory(account_factory)



"""Components for interacting with Regional Accounts, which are containers for RegionalAccount instances."""
from threading import RLock

from cachetools import cached, TTLCache
from zope import interface
from zope.component.factory import Factory
from zope.schema.fieldproperty import FieldProperty

from .account import account_factory
from .caching_key import aggregated_string_hash
from .interfaces import IRegionalAccounts
from .regional_account import regional_account_factory


@interface.implementer(IRegionalAccounts)
class RegionalAccounts:
    """Enumerable read-only mapping of RegionalAccount instances.

    Keys are AWS region strings and values are related
    cs.aws_account.regional_account.RegionalAccount instances.

    This implementation leverages the caching factory for the RegionalAccount
    value population.  This means that 2 separate RegionalAccounts objects will
    point to the same value reference for common factory call signatures.

    See README.md for usage.

    Dict Filter Spec:
        Partitions:
         aws: # valid AWS partition name.  If absent, defaults 'aws'
           IncludeNonRegional: True|False # include non-regional endpoint names, defaults to False
           Regions: #if absent, defaults to all available regions
             include: [list, of, regions] #if absent, defaults to all available regions
             exclude: [list, of, regions] #takes precedence over include

    Args:
        RateLimit: [see cs.ratelimit.components.ratelimitproperties_factory]
        Account: [see cs.aws_account.account.account_factory]

    Kwargs:
        Filter: Dict filter spec (see above)
        RateLimitRegionSpec: dict where keys are region names and values are
          cs.ratelimit.components.ratelimitproperties_factory factory specs
        service: boto3.session.Session.Client() service used as reference to
                 build region name lists.
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments, consider-using-generator, invalid-name

    filter = FieldProperty(IRegionalAccounts['filter'])

    def __init__(self, RateLimit, Account, Filter=None, RateLimitRegionSpec=None, service='ec2'):
        """Initialize the RegionalAccounts container."""
        self.filter = Filter if Filter else {}

        self._rate_limit = RateLimit
        self._account_properties = Account
        self._account = account_factory(**Account)
        self._service = service
        self._rate_limit_region_spec = RateLimitRegionSpec if RateLimitRegionSpec else {}

        self._rlock = RLock()
        self._regional_accounts = {}

    def _get_regional_account(self, region_name):
        with self._rlock:
            try:
                return self._regional_accounts[region_name]
            except KeyError:
                self._regional_accounts[region_name] = regional_account_factory(
                    self._rate_limit_region_spec[region_name] if region_name in
                    self._rate_limit_region_spec else self._rate_limit, self._account_properties, region_name)
            return self._regional_accounts[region_name]

    def account(self):
        """Return the accounts wrapped by this RegionalAccounts instance."""
        return self._account

    def __getitem__(self, key):
        """Get the requested regional account or raise as part of IEnumerableMapping."""
        if key not in self:
            raise KeyError(key)
        return self._get_regional_account(key)

    def get(self, key, default=None):
        """Get an attribute or return the default as part of IEnumerableMapping."""
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        """Return whether the requested regional account is held by this wrapper as part of IEnumerableMapping."""
        return key in self.__iter__()

    def keys(self):
        """Return the requested regional account key held by this wrapper as part of IEnumerableMapping."""
        return (k for k in self)

    @cached(cache=TTLCache(maxsize=1000, ttl=86400))
    def _all_regions(self, partition_name, allow_non_regional):
        return self.account().session().boto3().get_available_regions(
                                self._service,
                                partition_name=partition_name,
                                allow_non_regional=allow_non_regional)

    def __iter__(self):
        """Iterate over all regions in this wrapper as part of IEnumerableMapping."""
        partitions = self.filter.get('Partitions', {'aws': None})

        regions = set()
        for name in set(partitions):
            p_config = partitions[name]
            if not p_config:
                p_config = {}
            include_non_regional = p_config.get('IncludeNonRegional', False)
            all_regions = self._all_regions(name, include_non_regional)

            region_include = set(p_config.get('Regions', {}).get('include', {}))
            if not region_include:
                region_include = set(all_regions)
            region_exclude = set(p_config.get('Regions', {}).get('exclude', {}))
            regions.update(region_include - region_exclude)

        for region in regions:
            yield region

    def values(self):
        """Return all regional accounts in this wrapper as part of IEnumerableMapping."""
        return tuple([self[k] for k in self])

    def items(self):
        """Return key and account for all regional accounts in this wrapper as part of IEnumerableMapping."""
        return tuple([(k, self[k]) for k in self])

    def __len__(self):
        """Return the number of regional accounts held in this wrapper as part of IEnumerableMapping."""
        return len(list(iter(self)))


RegionalAccountsFactory = Factory(RegionalAccounts)


@interface.implementer(IRegionalAccounts)
@cached(cache={}, key=aggregated_string_hash, lock=RLock())
def regional_accounts_factory(**kwargs):
    """Create and cache a cs.aws_account.regional_accounts.RegionalAccounts instance.

    Common call signatures will return cached object.

    Create cs.aws_account.regional_accounts.RegionalAccounts with defined
    rate limits for Account and optional region filter.

    Kwargs:
        RateLimit: optional [see cs.ratelimit.components.ratelimitproperties_factory]
        Account: required [see cs.aws_account.account.account_factory]
        Filter: optional valid dict filter specification
        RateLimitRegionSpec: optional per-region rate limit specs
        service: optional boto3 client service reference to build region specs from

    Returns:
        cs.aws_account.regional_accounts.RegionalAccounts object
    """
    return RegionalAccounts(**kwargs)


CachingRegionalAccountsFactory = Factory(regional_accounts_factory)



import os
import threading
import time
import unittest

from zope import component
from zope.interface.verify import verifyObject

from ..interfaces import ISession
from ..testing import AWS_ACCOUNT_INTEGRATION_LAYER
from cs.aws_account.b3 import IBoto3Session
from cs.aws_account.session import Session, session_factory


class IntegrationTestAWSAccountSession(unittest.TestCase):

    level = 2

    def setUp(self):
        self.session_kwargs = {
            'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
            'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY'),
        }
        self.assume_role_kwargs = {
            'sts_method': 'assume_role',
            'RoleArn': os.environ.get('AWS_ASSUME_ROLE'),
            'RoleSessionName': 'testing_assume_role_for_cs_aws_account_package',
        }

    def test_simple_boto3(self):
        s = Session(**self.session_kwargs)
        self.assertTrue(ISession.providedBy(s))
        verifyObject(ISession, s)

        b3 = s.boto3()
        self.assertTrue(IBoto3Session.providedBy(b3))
        self.assertIs(b3, s.boto3())

    def test_get_credentials(self):
        lock = threading.Lock()
        b3_sessions = []
        s = Session(**self.session_kwargs)
        s.assume_role(**self.assume_role_kwargs)

        def add_session(s, b3_sessions):
            with lock:
                b3_sessions.append(s.boto3())

        t1 = threading.Thread(target=add_session, args=[s, b3_sessions])
        t2 = threading.Thread(target=add_session, args=[s, b3_sessions])
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        b3_sessions.append(s.boto3())
        self.assertIs(b3_sessions[0]._session._credentials,
                      b3_sessions[1]._session._credentials)
        self.assertIs(b3_sessions[1]._session._credentials,
                      b3_sessions[2]._session._credentials)

    def test_access_key(self):
        s = Session(**self.session_kwargs)
        self.assertEqual(self.session_kwargs['aws_access_key_id'],
                         s.access_key())

    def test_caches(self):
        s = Session(**self.session_kwargs)

        def get_cache_lengths():
            return (
                len(s._cache_access_key),
                len(s._cache_account_id),
                len(s._cache_user_id),
                len(s._cache_arn),
            )

        self.assertEqual((0, 0, 0, 0), get_cache_lengths())
        s.access_key()
        s.account_id()
        s.user_id()
        s.arn()
        self.assertEqual((1, 1, 1, 1), get_cache_lengths())

        s.access_key()
        s.account_id()
        s.user_id()
        s.arn()
        self.assertEqual((1, 1, 1, 1), get_cache_lengths())

    def test_cache_ttl(self):
        s = Session(cache_ttl=1, **self.session_kwargs)

        def get_cache_lengths():
            return (
                len(s._cache_access_key),
                len(s._cache_account_id),
                len(s._cache_user_id),
                len(s._cache_arn),
            )

        self.assertEqual((0, 0, 0, 0), get_cache_lengths())

        s.access_key()
        s.account_id()
        s.user_id()
        s.arn()
        self.assertEqual((1, 1, 1, 1), get_cache_lengths())

        time.sleep(1)
        self.assertEqual((0, 0, 0, 0), get_cache_lengths())

        s.access_key()
        s.account_id()
        s.user_id()
        s.arn()
        self.assertEqual((1, 1, 1, 1), get_cache_lengths())

    def test_threadlocal_boto3(self):
        s = Session(**self.session_kwargs)
        lock = threading.Lock()
        d = []
        errors = []

        def tl_session():
            try:
                with lock:
                    d.append(s.boto3())
            except Exception as e:
                errors.append(e)
                return

        t1 = threading.Thread(target=tl_session)
        t2 = threading.Thread(target=tl_session)
        t1.start()
        t2.start()
        t1.join()
        t2.join()
        self.assertEqual(errors, [])
        self.assertIsNot(d[0], d[1])
        self.assertEqual(d[0].get_credentials().access_key,
                         d[1].get_credentials().access_key)

    def test_assume_role_and_revert(self):
        s = Session(**self.session_kwargs)
        ak1 = s.access_key()
        s.assume_role(**self.assume_role_kwargs)
        self.assertNotEqual(ak1, s.access_key())
        s.revert()
        self.assertEqual(ak1, s.access_key())
        s.revert()
        self.assertEqual(ak1, s.access_key())

        s.assume_role(**self.assume_role_kwargs)
        s1 = s.boto3()
        s.revert()
        s.assume_role(**self.assume_role_kwargs)
        s2 = s.boto3()
        s.revert()
        s1.client('sts').get_caller_identity()['UserId']
        s2.client('sts').get_caller_identity()['UserId']

    def test_assume_role_deferrence(self):
        s = Session(**self.session_kwargs)
        ak1 = s.access_key()
        s.assume_role(deferred=True, **self.assume_role_kwargs)
        self.assertNotEqual(ak1, s.access_key())
        s.revert()
        self.assertEqual(ak1, s.access_key())
        s.revert()
        self.assertEqual(ak1, s.access_key())

        s.assume_role(deferred=True, **self.assume_role_kwargs)
        s1 = s.boto3()
        s.revert()
        s.assume_role(deferred=True, **self.assume_role_kwargs)
        s2 = s.boto3()
        s.revert()
        s1.client('sts').get_caller_identity()['UserId']
        s2.client('sts').get_caller_identity()['UserId']

    def test_threadlocal_assume_role(self):
        s = Session(**self.session_kwargs)
        ak1 = s.access_key()
        t1 = threading.Thread(target=s.assume_role, kwargs=self.assume_role_kwargs)
        t1.start()
        t1.join()
        self.assertNotEqual(ak1, s.access_key())

    def test_threadlocal_revert(self):
        s = Session(**self.session_kwargs)
        ak1 = s.access_key()
        s.assume_role(**self.assume_role_kwargs)
        ak2 = s.access_key()
        t1 = threading.Thread(target=s.revert)
        t1.start()
        t1.join()
        self.assertEqual(ak1, s.access_key())
        self.assertNotEqual(ak2, s.access_key())

    def test_caching_session_factory(self):
        s1 = session_factory(SessionParameters=self.session_kwargs)
        s2 = session_factory(SessionParameters=self.session_kwargs)
        self.assertIs(s1, s2)

        s1 = session_factory(SessionParameters=self.session_kwargs,
                             AssumeRole=self.assume_role_kwargs)
        s2 = session_factory(SessionParameters=self.session_kwargs,
                             AssumeRole=self.assume_role_kwargs)
        self.assertIs(s1, s2)

        assume_role_kwargs = self.assume_role_kwargs.copy()
        assume_role_kwargs['RoleSessionName'] = 'testing_assume_role_for_cs_aws_account_package_caching_test'
        s2 = session_factory(SessionParameters=self.session_kwargs,
                             AssumeRole=assume_role_kwargs)
        self.assertIsNot(s1, s2)


class IntegrationTestAWSAccountSessionZCA(unittest.TestCase):

    layer = AWS_ACCOUNT_INTEGRATION_LAYER

    def setUp(self):
        self.session_kwargs = {'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
                               'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY')}

    def test_session_factories(self):
        s = component.createObject(u"cs.aws_account.session", SessionParameters=self.session_kwargs)
        self.assertTrue(ISession.providedBy(s))
        s = component.createObject(u"cs.aws_account.cached_session", SessionParameters=self.session_kwargs)
        self.assertTrue(ISession.providedBy(s))

        s = component.createObject(u"cs.aws_account.cached_session")
        self.assertTrue(ISession.providedBy(s))


class TestSessionClientKwargs(unittest.TestCase):
    def setUp(self):
        self.session_kwargs = {
            'region_name': 'us-west-1',
            'ServiceEndpoints': {
                'sts': 'https://test.sts.us-west-1.com/test',
                'sqs': {
                    'us-west-1': 'https://test.sqs.us-west-1.com/test',
                    'eu-central-1': 'https://test.sqs.eu-central-1.com/test',
                }
            }
        }

    def test_with_service_and_session_region_matches_endpoint_region(self):
        s = Session(**self.session_kwargs)
        self.assertDictEqual(
            s.client_kwargs(service='sts'),
            {'region_name': 'us-west-1', 'endpoint_url': 'https://test.sts.us-west-1.com/test'}
        )

    def test_cross_region_service_endpoint(self):
        s = Session(**self.session_kwargs)
        self.assertDictEqual(
            s.client_kwargs(service='sqs'),
            {'region_name': 'us-west-1', 'endpoint_url': 'https://test.sqs.us-west-1.com/test'}
        )

    def test_no_endpoint_url_added_without_service(self):
        s = Session(**self.session_kwargs)
        self.assertDictEqual(s.client_kwargs(), {'region_name': 'us-west-1'})

    def test_no_endpoint_url_added_with_service_not_in_endpoints_map(self):
        s = Session(**self.session_kwargs)
        self.assertDictEqual(s.client_kwargs(service='invalid'), {'region_name': 'us-west-1'})

    def test_no_endpoint_url_added_with_service_not_in_cross_region_endpoints_map(self):
        self.session_kwargs['region_name'] = 'us-west-2'
        s = Session(**self.session_kwargs)
        self.assertDictEqual(s.client_kwargs(service='sqs'), {'region_name': 'us-west-2'})

    def test_no_endpoint_url_added_for_session_region_mismatch(self):
        self.session_kwargs['ServiceEndpoints']['sts'] = 'https://sts.ap-east-1.com'
        s = Session(**self.session_kwargs)
        self.assertDictEqual(s.client_kwargs(service='sts'), {'region_name': 'us-west-1'})

    def test_no_endpoint_added_without_session_region(self):
        self.session_kwargs.pop('region_name')
        s = Session(**self.session_kwargs)
        self.assertDictEqual(s.client_kwargs(service='sts'), {})



"""Monkey patching for various botocore features."""
from threading import RLock

from botocore.loaders import Loader
# pylint: disable=global-variable-not-assigned, invalid-name


_cache = {}
_lock = RLock()


def global_cache(func):
    """Cache the result of a method globally.

    This is essentially a copy of botocore.loaders.Loader.instance_cache but
    with a synchronized global cache (vs instance-specific)
    """
    def _wrapper(self, *args, **kwargs):
        global _cache, _lock
        with _lock:
            key = (func.__name__,) + args
            for pair in sorted(kwargs.items()):
                key += pair
            if key in _cache:
                return _cache[key]
            data = func(self, *args, **kwargs)
            _cache[key] = data
            return data
    return _wrapper


Loader_init_orig = Loader.__init__


def Loader_init_monkey(*args, **kwargs):
    """Wrap and monkey patch the original botocore Loader."""
    Loader_init_orig(*args, **kwargs)



"""Helpers for creating a caching key."""


def aggregated_string_hash(*args, **kwargs):
    """Return a hash based on aggregated object strings."""
    _id = ""
    for arg in args:
        _id += str(arg)

    if kwargs:
        keys = sorted(kwargs.keys())
        for k in keys:
            _id += f"{k}:{kwargs[k]}"
    return hash(_id)



from datetime import timedelta
import os
import unittest

from cs.ratelimit import RateLimitProperties, RateLimitExceeded
from zope import component

from ..account import Account
from ..interfaces import IRegionalAccount
from ..regional_account import RegionalAccount
from ..session import Session
from ..testing import AWS_ACCOUNT_INTEGRATION_LAYER


class IntegrationTestAWSAccountRegionalAccount(unittest.TestCase):

    level = 2

    def setUp(self):
        self.session_kwargs = {'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
                               'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY')}
        self.session = Session(**self.session_kwargs)
        self.account = Account(self.session)
        self.ratelimit = RateLimitProperties(max_count=1, interval=timedelta(seconds=10), block=False)
        self.ra = RegionalAccount(self.ratelimit, self.account, region_name='us-east-1')

    def test_region(self):
        self.assertEqual(self.ra.region(), 'us-east-1')

    def test_account(self):
        self.assertIs(self.ra.account(), self.account)

    def test_call_client(self):
        self.ra.call_client('sts', 'get_caller_identity')
        with self.assertRaises(RateLimitExceeded):
            self.ra.call_client('sts', 'get_caller_identity')
        self.ra.ratelimit.max_count = 2
        self.ra.call_client('sts', 'get_caller_identity')

    def test_get_paginator(self):
        paginator = self.ra.get_paginator('ec2', 'describe_instances')
        page_iterator = paginator.paginate(PaginationConfig={'MaxItems': 10})
        next(page_iterator.__iter__())
        with self.assertRaises(RateLimitExceeded):
            next(page_iterator.__iter__())
        self.ra.ratelimit.max_count = 2
        next(page_iterator.__iter__())


class IntegrationTestAWSAccountRegionalAccountZCA(unittest.TestCase):

    layer = AWS_ACCOUNT_INTEGRATION_LAYER

    def setUp(self):
        self.session_kwargs = {'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
                               'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY')}
        self.session = Session(**self.session_kwargs)
        self.account = Account(self.session)
        self.ratelimit = RateLimitProperties(max_count=1, interval=timedelta(seconds=1), block=False)

    def test_factories(self):
        # non-cached
        ra = component.createObject(
            u"cs.aws_account.regional_account", self.ratelimit, self.account, region_name='us-east-1')
        self.assertTrue(IRegionalAccount.providedBy(ra))

        # cached
        self.assertIs(
                component.createObject(u"cs.aws_account.cached_regional_account",
                                       Account={'SessionParameters': self.session_kwargs}),
                component.createObject(u"cs.aws_account.cached_regional_account",
                                       Account={'SessionParameters': self.session_kwargs})
            )



"""Declares interfaces for wiring components together."""
# pylint: disable=inherit-non-class, no-method-argument, no-self-argument, too-many-ancestors
from typing import Optional

from zope import interface
from zope import schema
from zope.interface.common.collections import IMutableMapping
from zope.interface.common.mapping import IEnumerableMapping

from cs import ratelimit


class ISession(interface.Interface):
    """Boto3 Session accessor."""

    def boto3():
        """Return active boto3 session."""

    def assume_role(sts_method='assume_role', deferred=False, **kwargs):
        """Replace active boto3 session with assumed role boto3 session.

        Kwargs:
            sts_method: name of sts Client method used to execute the role assumption
            deferred: Defers the actual role assumption operation until boto3() is called
            [others]:  Other kwargs are passed unalltered into the named sts role assumption method
        """

    def client_kwargs(service: Optional[str] = None) -> IMutableMapping:
        """Return shallow copy of self._client_kwargs.

        Args:
            service (Optional[str]): If given, and a service endpoint override
                is configured for the specified AWS service (e.g., 'sqs', 'sts'),
                it will be added to the returned client_kwargs mapping.

        Returns:
            mutable mapping of kwargs to pass to a `boto3.session.Session.client` call.
        """

    def revert():
        """Revert active session to previous boto3 session in used before last call to assume_role().

        Attempting to revert the originating boto3 session object does nothing.

        Returns
            active session previous to the revert operation
        """

    def access_key():
        """Return access key string in use for the referenced boto3 object."""

    def account_id():
        """Return AWS account identity string in use for the referenced boto3 object."""

    def user_id():
        """Return AWS account identity string in use for the referenced boto3 object."""

    def arn():
        """Return AWS account identity string in use for the referenced boto3 object."""


class IAccount(interface.Interface):
    """AWS account."""

    def account_id():
        """Return AWS account identity string."""

    def alias():
        """Return first AWS account alias string else account identity."""

    def aliases():
        """Return iterable of all AWS account alias strings."""

    def session():
        """Return ISession provider."""


class IRegionalAccount(interface.Interface):
    """A boto3 session client caller with rate limiting capabilities."""

    ratelimit = schema.Object(
            title="Rate limit properties",
            description="Rate limiting properties",
            required=True,
            schema=ratelimit.IRateLimitProperties
        )

    def account():
        """Return IAccount provider."""

    def region():
        """Return AWS region string."""

    def call_client(self, service, method, client_kwargs=None, **kwargs):
        """Return call to boto3 service client method limited by properties in ratelimit.

        This can raise cs.ratelimit.RateLimitExceeded based on ratelimit settings.

        Raises:
            [dependent on named boto3 service method]

        Args:
            service: valid boto3 service name string
            method: valid boto3 named service method name string
            client_kwargs: mapping of kwargs that will be used to create the
                boto3.client object.

        Kwargs:
            [dependent on named boto3 service method]

        Returns:
            [dependent on named boto3 service method]
        """

    def get_paginator(service, method, client_kwargs=None, **kwargs):
        """Return paginator for boto3 service client method limited by properties in ratelimit.

        Same call features as call_client() except calls are accessed via
        a returned paginator.
        """


class IRegionalAccounts(IEnumerableMapping):
    """Mapping whose keys are AWS region strings and values are related IRegionalAccount providers."""

    filter = schema.Dict(
            title="Filter",
            description="The container filter specification",
            readonly=True,  # but still mutable
            required=True
        )

    def account():
        """Return IAccount provider."""


class IRegionalAccountSet(interface.Interface):
    """A container of IRegionalAccounts providers that iterates over their IRegionalAccount providers."""

    def add(regional_accounts):
        """Add IRegionalAccounts provider to include for iteration if not available."""

    def discard(regional_accounts):
        """Discard IRegionalAccounts provider from iteration if available."""

    def values():
        """Frozenset of available IRegionalAccounts providers."""

    def __iter__():
        """Iterate unique IRegionalAccount providers from available IRegionalAccounts providers."""



"""Provides an optimized thread-safe boto3 session wrapper."""
# pylint: disable=invalid-name
from collections.abc import Mapping
from threading import local, RLock
from typing import Optional
import logging
import operator

from boto3.session import Session as botoSession
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session
from cachetools import cached, cachedmethod, Cache
from zope import interface
from zope.component.factory import Factory
from zope.interface.common.collections import IMutableMapping

from .caching_key import aggregated_string_hash
from .interfaces import ISession
from .retry import aws_throttling_retry

logger = logging.getLogger(__name__)


# Region list as of 2022-10-27. Note: A more future proof solution would be
# to leverage boto3.get_available_regions, but we do this instead to prevent
# additional API calls.
AWS_REGIONS = set([
    'af-south-1', 'ap-east-1', 'ap-northeast-1', 'ap-northeast-2', 'ap-northeast-3',
    'ap-south-1', 'ap-southeast-1', 'ap-southeast-2', 'ap-southeast-3', 'ca-central-1',
    'eu-central-1', 'eu-north-1', 'eu-south-1', 'eu-west-1', 'eu-west-2', 'eu-west-3',
    'me-south-1', 'sa-east-1', 'us-east-1', 'us-east-2', 'us-west-1', 'us-west-2',
    'us-gov-east-1', 'us-gov-west-1'
])


@interface.implementer(ISession)
class Session:
    """Thread-safe boto3 Session accessor.

    Provides convenient access to boto3.session.Session() objects even in
    complex threaded environments with IAM role assumption.

    Some method calls have memoizing caches to improve performance of
    common action calls (such as logging session information).

    Kwargs:
        [boto3.session.Session]: See boto3.session.Session for available kwargs
    """

    # pylint: disable=too-many-instance-attributes, protected-access

    def _reset_caches(self):
        """Reset the memoizing decorator caches."""
        with self._rlock:
            self._cache_access_key = Cache(maxsize=1)
            self._cache_account_id = Cache(maxsize=1)
            self._cache_user_id = Cache(maxsize=1)
            self._cache_arn = Cache(maxsize=1)

    def __init__(self, **SessionParameters):
        """Set up the thread safety and threadlocal data."""

        class TLBoto3(local):  # pylint: disable=too-few-public-methods
            """Threadlocal boto3 client stack."""

            boto3 = []  # threadlocal stack

        # We can't use the ServiceEndpoints directly in boto3 session creation,
        # so we apply them to each client, depending on which service the
        # client is for, instead of adding them to the `_stack`.
        self._service_endpoints = SessionParameters.pop('ServiceEndpoints', {})

        self._local = TLBoto3()  # threadlocal data to protect the non-TS low-level Boto3 session
        self._stack = [(aggregated_string_hash(SessionParameters), SessionParameters)]  # master stack
        self._rlock = RLock()
        self._client_kwargs = {}
        self._credentials = {}
        if 'region_name' in SessionParameters:
            self._client_kwargs['region_name'] = SessionParameters['region_name']
        self._reset_caches()

    def _get_credentials(self, boto3_session, sts_method='assume_role', **kwargs):
        """Efficient Boto3 credentials getter.

        Only one botocore.credentials.RefreshableCredentials object should
        be needed per assume-role based boto3 session objects.  To prevent
        un-needed api calls for threaded usage, we'll syncronize access to an
        object-level storage.

        Same args as _assume_role().
        """
        # botocore.credentials.RefreshableCredentials is seemingly thread-safe
        hash_ = boto3_session._aws_account_hash
        role_id = "".join([str(v) for v in kwargs.values()])
        with self._rlock:
            if hash_ not in self._credentials:
                self._credentials[hash_] = {}
            if sts_method not in self._credentials[hash_]:
                self._credentials[hash_][sts_method] = {}
            if role_id not in self._credentials[hash_][sts_method]:
                self._credentials[hash_][sts_method][role_id] = None

            if not self._credentials[hash_][sts_method][role_id]:

                def refresh():
                    logger.debug(
                        "Refreshing assumed role credentials for ARN %s with session name %s",
                        kwargs.get('RoleArn', ''), kwargs.get('RoleSessionName', ''))
                    assume_role = getattr(boto3_session.client('sts', **self.client_kwargs(service='sts')), sts_method)
                    logger.debug("Attempting to assumed AWS Role with data %s", kwargs)
                    credentials = assume_role(**kwargs)['Credentials']
                    logger.info("Assumed AWS Role with data %s", kwargs)
                    # mapping keys common among all assume_role call variations
                    return {
                        'access_key': credentials['AccessKeyId'],
                        'secret_key': credentials['SecretAccessKey'],
                        'token': credentials['SessionToken'],
                        'expiry_time': credentials['Expiration'].isoformat(),
                    }

                session_credentials = RefreshableCredentials.create_from_metadata(
                    metadata=refresh(),
                    refresh_using=refresh,
                    method=f"sts-{sts_method.replace('_', '-')}")  # assume_role -> sts-assume-role
                self._credentials[hash_][sts_method][role_id] = session_credentials

            return self._credentials[hash_][sts_method][role_id]

    def _assume_role(self, boto3_session, sts_method='assume_role', **kwargs):
        """Stateless assume role call; returns Boto3 session with role assumption."""
        # https://programtalk.com/python-examples/botocore.credentials.RefreshableCredentials.create_from_metadata/
        session = get_session()
        session._credentials = self._get_credentials(boto3_session, sts_method, **kwargs)
        return botoSession(botocore_session=session)

    def boto3(self):
        """Return threadlocal active boto3.session.Session object."""
        with self._rlock:
            # make sure threadlocal boto3 stack entries are valid
            for i, _hash, _ in [(i, b3[0], b3[1],) for i, b3 in enumerate(self._local.boto3)]:
                try:
                    if _hash != self._stack[i][0]:
                        self._local.boto3 = self._local.boto3[:i]
                        break
                except IndexError:
                    self._local.boto3 = self._local.boto3[:i]
                    break

            # reduce threadlocal boto3 stack, if needed
            self._local.boto3 = self._local.boto3[:len(self._stack)]
            # rebuild the stack so we can safely reference info in unlocked state.
            stack = [t for t in self._stack[len(self._local.boto3):]]  # pylint: disable=unnecessary-comprehension

        # Do these tasks outside the lock to allow concurrent calls
        # to independently build the threadlocal boto3 object
        # logger.debug("Preparing threadlocal stack")
        for hash_, kwargs in stack:
            if not self._local.boto3:
                boto_session = botoSession(**kwargs)
                boto_session._aws_account_hash = hash_  # mark the object with its hash
                self._local.boto3.append((hash_, boto_session,))
            else:
                boto_session = self._assume_role(self._local.boto3[-1][1], **kwargs)
                boto_session._aws_account_hash = hash_  # mark the object with its hash
                self._local.boto3.append((hash_, boto_session,))
        boto_session = self._local.boto3[-1][1]  # return lifo entry from threadlocal stack
        logger.debug("Returning Boto3.Session object with access key %s", boto_session.get_credentials().access_key)
        return boto_session

    def revert(self):
        """Set active boto3.session.Session object to previous; return pop'd threadlocal boto3.session.Session."""
        with self._rlock:
            if len(self._stack) > 1:
                boto_session = self.boto3()
                self._stack.pop()
                self._reset_caches()
                logger.debug("Reverting role to %s", self._stack[-1])
                return boto_session
        return None

    def assume_role(self, sts_method='assume_role', deferred=False, **kwargs):
        """Set active boto3.session.Session object to role-assumed object."""
        with self._rlock:
            kwargs['sts_method'] = sts_method
            self._stack.append((aggregated_string_hash(kwargs), kwargs,))
            self._reset_caches()
        if not deferred:
            self.boto3()  # init, raises on error
        else:
            logger.info("Differing assumption of AWS Role with args %s", kwargs)

    def client_kwargs(self, service: Optional[str] = None) -> IMutableMapping:
        """Return shallow copy of self._client_kwargs.

        Args:
            service (Optional[str]): If given, and a service endpoint override
                is configured for the specified AWS service (e.g., 'sqs', 'sts'),
                it will be added to the return client_kwargs dict.
        """
        client_kwargs = self._client_kwargs.copy()
        kwarg_region = client_kwargs.get('region_name')
        if service and kwarg_region:
            endpoint_url = self._service_endpoints.get(service)

            # handle cross-region custom endpoints
            if isinstance(endpoint_url, Mapping):
                endpoint_url = endpoint_url.get(kwarg_region, None)

            # if we have a custom endpoint for this region
            if endpoint_url:
                endpoint_region = None
                endpoint_url_parts = endpoint_url.split('.')
                for part in endpoint_url_parts:
                    if part in AWS_REGIONS:
                        endpoint_region = part
                # Iff the custom endpoint region and operation match, use the endpoint for that region.
                # Otherwise, we have no choice but to fallback to the default endpoint, since boto3
                # requires a region if endpoint_url is also given.
                if endpoint_region and endpoint_region == kwarg_region:
                    client_kwargs['endpoint_url'] = endpoint_url
        return client_kwargs

    @cachedmethod(operator.attrgetter('_cache_access_key'), lock=operator.attrgetter('_rlock'))
    @aws_throttling_retry()
    def access_key(self):
        """Return access key related to session."""
        logger.debug("Refreshing access key cache")
        creds = self.boto3().get_credentials()  # returns None if not authenticated
        return creds.access_key if creds else None

    @cachedmethod(operator.attrgetter('_cache_account_id'), lock=operator.attrgetter('_rlock'))
    @aws_throttling_retry()
    def account_id(self):
        """Return AWS account ID related to session."""
        logger.debug("Refreshing account id cache")
        return self.boto3().client('sts', **self.client_kwargs(service='sts')).get_caller_identity()['Account']

    @cachedmethod(operator.attrgetter('_cache_user_id'), lock=operator.attrgetter('_rlock'))
    @aws_throttling_retry()
    def user_id(self):
        """Return AWS user ID related to session."""
        logger.debug("Refreshing user id cache")
        return self.boto3().client('sts', **self.client_kwargs(service='sts')).get_caller_identity()['UserId']

    @cachedmethod(operator.attrgetter('_cache_arn'), lock=operator.attrgetter('_rlock'))
    def arn(self):
        """Return the AWS Arn related to session (includes user name)."""
        # this call can be region dependent.  e.g. if calling aws from a govcloud
        # acct, this would fail because aws doesn't understand accounts in govcloud.
        logger.debug("Refreshing arn cache")
        return self.boto3().client('sts', **self.client_kwargs(service='sts')).get_caller_identity()['Arn']


SessionFactory = Factory(Session)


@interface.implementer(ISession)
@cached(cache={}, key=aggregated_string_hash, lock=RLock())
def session_factory(SessionParameters=None, AssumeRole=None, AssumeRoles=None):
    """Create and cache a cs.aws_account.session.Session.

    Common call signatures will return cached object.

    Create cs.aws_account.session.Session object. If AssumeRole parameter is
    available, then process the role assumption. If AssumeRoles parameter
    is available, then process the series of role assumptions.

    Kwargs:
        SessionParameters: optional [see cs.aws_account.session.Session]
        AssumeRole: optional [see cs.aws_account.session.Session.assume_role]
        AssumeRoles: optional iterable of assume_role mappings

    Returns:
        cs.aws_account.session.Session object
    """
    SessionParameters = SessionParameters if SessionParameters else {}
    session = Session(**SessionParameters)
    if AssumeRole:
        session.assume_role(**AssumeRole)
    if AssumeRoles:
        for Role in AssumeRoles:
            session.assume_role(**Role)
    return session


CachingSessionFactory = Factory(session_factory)



import os
import unittest

from zope import component

from ..testing import AWS_ACCOUNT_INTEGRATION_LAYER
from ..interfaces import IAccount
from ..account import Account, account_factory
from ..session import Session


class IntegrationTestAWSAccountAccount(unittest.TestCase):

    level = 2

    def setUp(self):
        self.session_kwargs = {'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
                               'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY')}
        self.assume_role_kwargs = {'sts_method': 'assume_role',
                                   'RoleArn': os.environ.get('AWS_ASSUME_ROLE'),
                                   'RoleSessionName': 'testing_assume_role_for_cs_aws_account_package'
                                   }
        self.session = Session(**self.session_kwargs)

    def test_account_id(self):
        acct = Account(self.session)
        self.assertGreater(len(acct.account_id()), 0)

    def test_account_alias(self):
        acct = Account(self.session)
        self.assertGreater(len(acct.alias()), 0)

    def test_account_aliases(self):
        acct = Account(self.session)
        self.assertEqual(len(acct._cache_aliases), 0)
        self.assertGreater(len(acct.aliases()), 0)
        self.assertEqual(len(acct._cache_aliases), 1)

    def test_account_session(self):
        acct = Account(self.session)
        self.assertIs(acct.session(), self.session)

    def test_caching_factory(self):
        s1 = account_factory(SessionParameters=self.session_kwargs)
        s2 = account_factory(SessionParameters=self.session_kwargs)
        self.assertIs(s1, s2)


class IntegrationTestAWSAccountAccountZCA(unittest.TestCase):
    layer = AWS_ACCOUNT_INTEGRATION_LAYER

    def setUp(self):
        self.session_kwargs = {'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
                               'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY')}

    def test_account_factories(self):
        s = component.createObject(u"cs.aws_account.session", SessionParameters=self.session_kwargs)
        a = component.createObject(u"cs.aws_account.account", s)
        self.assertTrue(IAccount.providedBy(a))
        a1 = component.createObject(u"cs.aws_account.cached_account", SessionParameters=self.session_kwargs)
        self.assertTrue(IAccount.providedBy(a1))
        a2 = component.createObject(u"cs.aws_account.cached_account", SessionParameters=self.session_kwargs)
        self.assertIs(a1, a2)



"""Components for interacting with RegionalAccountSets, which are containers for RegionalAccounts instances."""
from threading import RLock

from zope import interface
from zope.component.factory import Factory

from .interfaces import IRegionalAccounts, IRegionalAccountSet
from .regional_accounts import regional_accounts_factory


@interface.implementer(IRegionalAccountSet)
class RegionalAccountSet:
    """A container of cs.aws_account.regional_accounts.RegionalAccounts instances.

    The container iterates over its content values
    (cs.aws_account.regional_account.RegionalAccount objects).

    add(), discard() and values() calls refer to RegionAccounts containers,
    whilst __iter__() returns RegionAccount objects.

    Args:
        arbitrary list of RegionalAccounts objects to instantiate the active set with
    """

    # pylint: disable=no-value-for-parameter
    def __init__(self, *args):
        """Initialize the RegionalAccountSet."""
        self._regional_accounts = set()
        self._lock = RLock()
        for regional_account in args:
            if not IRegionalAccounts.providedBy(regional_account):
                raise ValueError(regional_account)
            self._regional_accounts.add(regional_account)

    def add(self, regional_accounts):
        """Add RegionalAccounts instance to include for iteration if not available."""
        if not IRegionalAccounts.providedBy(regional_accounts):
            raise ValueError(regional_accounts)
        with self._lock:
            self._regional_accounts.add(regional_accounts)

    def discard(self, regional_accounts):
        """Discard RegionalAccounts instance from iteration if available."""
        if not IRegionalAccounts.providedBy(regional_accounts):
            raise ValueError(regional_accounts)
        with self._lock:
            self._regional_accounts.discard(regional_accounts)

    def values(self):
        """Frozenset of available RegionalAccounts providers."""
        with self._lock:
            return frozenset(list(self._regional_accounts))

    def __iter__(self):
        """Iterate over unique RegionalAccount instances from available RegionalAccounts instances."""
        regional_accounts = []
        with self._lock:
            for regional_account in self._regional_accounts:
                regional_accounts.extend(list(regional_account.values()))
            regional_accounts = set(regional_accounts)
        return iter(regional_accounts)


RegionalAccountSetFactory = Factory(RegionalAccountSet)


@interface.implementer(IRegionalAccountSet)
def regional_account_set_factory(*args):
    """cs.aws_account.regional_account_set.RegionalAccountSet factory.

    This accepts an arbitrary number of valid
    cs.aws_account.regional_accounts.regional_accounts_factory kwargs dicts.
    the returned RegionalAccountSet will be instantiated with cached objects
    created with those specifications.

    Args:
        [see cs.aws_account.regional_accounts.regional_accounts_factory]

    Returns:
        cs.aws_account.regional_account_set.RegionalAccountSet object
    """
    ra_set = RegionalAccountSet()
    for regional_accounts in args:
        ra_set.add(regional_accounts_factory(**regional_accounts['RegionalAccounts']))
    return ra_set


RegionalAccountSetFactoryFromConfig = Factory(regional_account_set_factory)



"""A set of tools for working with AWS accounts.

The methods list_available_services() and list_api_versions() leverage
os.listdir() which is very slow when called from within
concurrent.futures.ThreadPoolExecutor instances.  Sadly, these are called when
a boto3.Session.create_client() is called.  This process becomes prohibitive
when the use-case creates new boto3 session objects in threads (like we do).

We'll monkey patch these two services to leverage a global call cache (vs the
default instance call cache), which insures a caches across instances.
"""

# pylint: disable=wrong-import-position

import botocore.loaders

from .monkies import global_cache


botocore.loaders.Loader.list_available_services = global_cache(
    botocore.loaders.Loader.list_available_services)
botocore.loaders.Loader.list_api_versions = global_cache(
    botocore.loaders.Loader.list_api_versions)


from .interfaces import IAccount
from .interfaces import IRegionalAccount
from .interfaces import IRegionalAccountSet
from .interfaces import IRegionalAccounts
from .interfaces import ISession
from .session import Session
from .session import session_factory  # caching
from .account import Account
from .account import account_factory  # caching
from .regional_account import RegionalAccount
from .regional_account import regional_account_factory  # caching
from .regional_accounts import RegionalAccounts
from .regional_accounts import regional_accounts_factory  # caching
from .regional_account_set import RegionalAccountSet
from .regional_account_set import regional_account_set_factory  # non-caching (but contained objects are cached)


__all__ = [
    'IAccount', 'IRegionalAccount', 'IRegionalAccountSet', 'IRegionalAccounts', 'ISession',
    'Session', 'session_factory', 'Account', 'account_factory', 'RegionalAccount', 'regional_account_factory',
    'RegionalAccounts', 'regional_accounts_factory', 'RegionalAccountSet', 'regional_account_set_factory',
]



import os
import unittest

from zope import component

from ..account import Account
from ..interfaces import IRegionalAccountSet
from ..regional_account import RegionalAccount
from ..regional_account_set import RegionalAccountSet
from ..regional_accounts import RegionalAccounts, regional_accounts_factory
from ..session import Session
from ..testing import AWS_ACCOUNT_INTEGRATION_LAYER


class IntegrationTestAWSAccountRegionalAccountSet(unittest.TestCase):

    level = 2

    def setUp(self):
        self.session_kwargs = {
            'aws_access_key_id':     os.environ.get('AWS_ACCESS_KEY_ID'),
            'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY'),
        }
        self.session = Session(**self.session_kwargs)
        self.account = Account(self.session)
        self.ratelimit_kwargs = {'max_count': 1, 'interval': 1, 'block': False}
        self.filter1 = {
            'Partitions': {
                'aws': {
                    'Regions': {'include': ['us-east-1']},
                },
            },
        }
        self.filter2 = {
            'Partitions': {
                'aws': {
                    'Regions': {'include': ['us-east-2']},
                },
            },
        }

        self.filter1_raccts = RegionalAccounts(
            self.ratelimit_kwargs, {'SessionParameters': self.session_kwargs}, self.filter1)
        self.filter2_raccts = RegionalAccounts(
            self.ratelimit_kwargs, {'SessionParameters': self.session_kwargs}, self.filter2)
        self.all_raccts = RegionalAccounts(self.ratelimit_kwargs, {'SessionParameters': self.session_kwargs})
        self.racct_set = RegionalAccountSet(self.filter1_raccts, self.filter2_raccts)

    def test_init_add_remove_values(self):
        self.assertEqual(len(self.racct_set.values()), 2)
        self.assertTrue(isinstance(list(self.racct_set.values())[0], RegionalAccounts))
        self.assertIn(self.filter1_raccts, self.racct_set.values())
        self.assertIn(self.filter2_raccts, self.racct_set.values())

        self.racct_set.discard(self.filter2_raccts)
        self.assertIn(self.filter1_raccts, self.racct_set.values())
        self.assertNotIn(self.filter2_raccts, self.racct_set.values())
        self.racct_set.discard(self.filter2_raccts)  # no error

        self.racct_set.add(self.filter2_raccts)
        self.assertIn(self.filter1_raccts, self.racct_set.values())
        self.assertIn(self.filter2_raccts, self.racct_set.values())
        self.racct_set.add(self.filter2_raccts)  # no error

    def test_iter(self):
        self.assertEqual(len(list(self.racct_set)), 2)
        self.assertTrue(isinstance(list(self.racct_set)[0], RegionalAccount))
        self.assertTrue(isinstance(list(self.racct_set)[1], RegionalAccount))
        self.assertIn(self.filter1_raccts['us-east-1'], self.racct_set)
        self.assertIn(self.filter2_raccts['us-east-2'], self.racct_set)
        self.assertNotIn(self.all_raccts['us-west-1'], self.racct_set)

        self.racct_set.discard(self.filter2_raccts)
        self.assertIn(self.filter1_raccts['us-east-1'], self.racct_set)
        self.assertNotIn(self.filter2_raccts['us-east-2'], self.racct_set)
        self.assertEqual(len(list(self.racct_set)), 1)

        self.racct_set.add(self.all_raccts)
        self.assertEqual(len(list(self.racct_set)), len(self.all_raccts))  # tests redundant removals


class IntegrationTestAWSAccountRegionalAccountSetZCA(unittest.TestCase):

    layer = AWS_ACCOUNT_INTEGRATION_LAYER

    def setUp(self):

        self.args = [
            {
                'RegionalAccounts': {
                    'RateLimit': {'max_count': 1, 'interval': 1, 'block': False},
                    'Account':  {
                        'SessionParameters': {
                            'aws_access_key_id': os.environ.get('AWS_ACCESS_KEY_ID'),
                            'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY'),
                        },
                    },
                    'Filter': {
                        'Partitions': {
                            'aws': {'Regions': {'include': ['us-east-1']}},
                        },
                    },
                },
            },
            {
                'RegionalAccounts': {
                    'RateLimit': {'max_count': 1, 'interval': 1, 'block': False},
                    'Account':  {
                        'SessionParameters': {
                            'aws_access_key_id': os.environ.get('AWS_ACCESS_KEY_ID'),
                            'aws_secret_access_key': os.environ.get('AWS_SECRET_ACCESS_KEY'),
                        },
                    },
                    'Filter': {
                        'Partitions': {
                            'aws': {'Regions': {'include': ['us-east-2']}},
                        },
                    },
                },
            },
        ]

    def test_account_factories(self):
        # non-cached
        ra0 = regional_accounts_factory(**self.args[0]['RegionalAccounts'])
        ra1 = regional_accounts_factory(**self.args[1]['RegionalAccounts'])
        ra_set = component.createObject(u"cs.aws_account.regional_account_set", ra0, ra1)
        self.assertTrue(IRegionalAccountSet.providedBy(ra_set))

        self.assertIsNot(
                component.createObject(u"cs.aws_account.regional_account_set_from_config",
                                       *self.args),
                component.createObject(u"cs.aws_account.regional_account_set_from_config",
                                       *self.args)
            )



from setuptools import setup, find_packages

version = '1.3.3'

tests_require = [
    'zope.testrunner',
    'zope.configuration',
    'bandit',
    'coverage',
]


doc_tests_require = [
    'pyyaml',
    'Jinja2',
]


lint_require = [
    'flake8',
    'pydocstyle',
    'pylint',
]


with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()


setup(name='cs.aws_account',
      version=version,
      description="AWS account components",
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Framework :: Zope",
        "Framework :: Flake8",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
      ],
      author='David Davis',
      author_email='david.davis@crowdstrike.com',
      maintainer='Forrest Aldridge',
      maintainer_email='forrest.aldridge@crowdstrike.com',
      url='https://github.com/CrowdStrike/cs.aws_account',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['cs'],
      include_package_data=True,
      package_data={
          '': ['*.zcml', '*.yaml', '*.md']
      },
      zip_safe=False,
      install_requires=[
          'setuptools',
          'zope.security',  # for zcml processing dep
          'zope.component',
          'zope.interface',
          'zope.schema',
          'boto3',
          'botocore',
          'cachetools',
          'cs.ratelimit'
      ],
      extras_require={
            'dev': tests_require + doc_tests_require + lint_require,
      },
      entry_points={},
      scripts=['util/run-tests.sh', 'util/lint.sh'],
      python_requires='>=3.8',
      )



"""Components for interacting with individual AWS Accounts by region."""
from threading import RLock
import logging
import operator
import types

from cachetools import cached
from zope import interface
from zope.component.factory import Factory
from zope.schema.fieldproperty import FieldProperty

from cs.ratelimit import ratelimitedmethod, ratelimitproperties_factory

from .retry import aws_throttling_retry
from .account import account_factory, Account as AwsAccount
from .caching_key import aggregated_string_hash
from .exceptions import AWSClientException
from .interfaces import IRegionalAccount


logger = logging.getLogger(__name__)


@interface.implementer(IRegionalAccount)
class RegionalAccount:
    """A   3 session client caller with rate limiting capabilities.

    Args:
        ratelimit: cs.ratelimit.RateLimitProperties instance that is referenced
                   at call time to determine rate limiting actions to call_client()
                   and the paginator returned by get_paginator()
        account: cs.aws_account.Account leveraged for all boto3 calls
        region_name: Valid string region_name for all boto3 calls
    """

    ratelimit = FieldProperty(IRegionalAccount['ratelimit'])  # cs.ratelimit.RateLimitProperties instance

    def __init__(self, ratelimit, account, region_name=None):
        """Initialize the RegionalAccount instance."""
        self.ratelimit = ratelimit
        self._account = account
        self._region_name = region_name

    def region(self):
        """Return referenced boto3 region_name string."""
        return self._region_name

    def account(self):
        """Return referenced cs.aws_account.Account instance."""
        return self._account

    def _get_client(self, service, **kwargs):
        kwargs['service_name'] = service
        kwargs['region_name'] = self.region()
        kwargs.update(self.account().session().client_kwargs(service=service))
        kwargs['config'] = AwsAccount.aws_client_config
        return self.account().session().boto3().client(**kwargs)

    @ratelimitedmethod(operator.attrgetter('ratelimit'))
    @aws_throttling_retry()
    def _limited(self, callback, **kwargs):
        try:
            logger.debug("calling AWS method %s for account %s (%s) region %s with user arn %s",
                         callback,
                         self._account.account_id(),
                         self._account.alias(),
                         self._region_name,
                         self._account.session().arn())
            return callback(**kwargs)
        except Exception as exc:
            raise AWSClientException(
                exc, AccountAlias=self._account.alias(), Region=self._region_name,
                AccountId=self._account.account_id()) from exc

    @aws_throttling_retry()
    def call_client(self, service, method, client_kwargs=None, **kwargs):
        """Return call to boto3 service client method limited by properties in ratelimit.

        This can raise cs.ratelimit.RateLimitExceeded based on ratelimit settings

        Raises:
            [dependent on named boto3 service method]

        Args:
            service: valid boto3 service name string
            method: valid boto3 named service method name string
            client_kwargs: mapping of kwargs that will be used to create the
                boto3.client object.

        Kwargs:
            [dependent on named boto3 service method]

        Returns:
            [dependent on named boto3 service method]
        """
        client_kwargs = {} if not client_kwargs else client_kwargs
        client = self._get_client(service, **client_kwargs)
        return self._limited(getattr(client, method), **kwargs)  # can raise AWSClientException

    def get_paginator(self, service, method, client_kwargs=None):
        """Return paginator for boto3 service client method limited by properties in ratelimit.

        Same call features as call_client() except calls are accessed via
        a returned paginator
        """
        client_kwargs = {} if not client_kwargs else client_kwargs
        _limited = self._limited

        def paginate(self, **kwargs):
            page_iterator = self.__wrapped__(**kwargs)  # get the default paginator from boto3
            _orig_method = page_iterator._method  # pylint: disable=protected-access

            def _method(_, **kwargs):
                return _limited(_orig_method, **kwargs)

            # over-ride with rate-limited method.
            page_iterator._method = types.MethodType(_method, page_iterator)  # pylint: disable=protected-access
            return page_iterator

        client = self._get_client(service, **client_kwargs)
        paginator = client.get_paginator(method)
        paginator.__wrapped__ = paginator.paginate  # we're gonna replace this
        paginator.paginate = types.MethodType(paginate, paginator)
        return paginator


RegionalAccountFactory = Factory(RegionalAccount)


@interface.implementer(IRegionalAccount)
@cached(cache={}, key=aggregated_string_hash, lock=RLock())
def regional_account_factory(RateLimit=None, Account=None, region_name=None):  # pylint: disable=invalid-name
    """Create a cached cs.aws_account.regional_account.RegionalAccount.

    Common call signatures will return cached object.

    Create cs.aws_account.regional_account.RegionalAccount with defined
    rate limits for Account/region_name combination.

    Kwargs:
        RateLimit: [see cs.ratelimit.components.ratelimitproperties_factory]
        Account: [see cs.aws_account.account.account_factory]
        region_name: valid AWS region name string (i.e. us-east-1, us-east-2, etc)

    Returns:
        cs.aws_account.regional_account.RegionalAccount object
    """
    RateLimit = RateLimit if RateLimit else {}
    rl_properties = ratelimitproperties_factory(**RateLimit)
    acct = account_factory(**Account)
    return RegionalAccount(rl_properties, acct, region_name=region_name)


CachingRegionalAccountFactory = Factory(regional_account_factory)
